@@ -1,25 +1,37 @@
 import math
+from abc import ABC
+from typing import Callable
 
 import torch
 from transformers import CLIPTokenizer, CLIPTextModel
 
 __all__ = ["EmbeddingsProvider"]
 
+class BaseTextualInversionManager(ABC):
+    pass
+
+
 class EmbeddingsProvider:
 
-    def __init__(self, tokenizer: CLIPTokenizer, text_encoder: CLIPTextModel):
+    def __init__(self,
+                tokenizer: CLIPTokenizer, # converts strings to lists of int token ids
+                text_encoder: CLIPTextModel, # convert a list of int token ids to a tensor of embeddings
+                textual_inversion_manager: BaseTextualInversionManager = None,
+                dtype_for_device_getter: Callable[[torch.device], torch.dtype] = lambda device: torch.float32,
+                ):
         self.tokenizer = tokenizer
         self.text_encoder = text_encoder
-        # self.textual_inversion_manager = TextualInversionManager(tokenizer=tokenizer, text_encoder=text_encoder)
+        self.textual_inversion_manager = textual_inversion_manager
+
+        # by default always use float32
+        self.get_dtype_for_device = dtype_for_device_getter
 
-    @property
-    def device(self):
-        return self.text_encoder.device
 
     @property
     def max_token_count(self) -> int:
         return self.tokenizer.model_max_length
 
+
     @classmethod
     def apply_embedding_weights(cls, embeddings: torch.Tensor, per_embedding_weights: list[float],
                                 normalize: bool) -> torch.Tensor:
@@ -32,9 +44,12 @@ class EmbeddingsProvider:
         # blended_embeddings now has shape (77, 768)
         return blended_embeddings
 
-    def get_embeddings_for_weighted_prompt_fragments(self, text_batch: list[list[str]],
+    def get_embeddings_for_weighted_prompt_fragments(self,
+                                                     text_batch: list[list[str]],
                                                      fragment_weights_batch: list[list[float]],
-                                                     should_return_tokens: bool = False) -> torch.Tensor:
+                                                     should_return_tokens: bool = False,
+                                                     device='cpu',
+                                 ) -> torch.Tensor | tuple[torch.Tensor, torch.Tensor]:
         """
 
         :param text_batch: A list of fragments of text to which different weights are to be applied.
@@ -62,7 +77,7 @@ class EmbeddingsProvider:
             # closer the resulting embedding is to an embedding for a prompt that simply lacks this fragment.
 
             # handle weights >=1
-            tokens, per_token_weights = self.get_token_ids_and_expand_weights(fragments, weights)
+            tokens, per_token_weights = self.get_token_ids_and_expand_weights(fragments, weights, device=device)
             base_embedding = self.build_weighted_embedding_tensor(tokens, per_token_weights)
 
             # this is our starting point
@@ -79,10 +94,9 @@ class EmbeddingsProvider:
             # such that the resulting lerped embedding is exactly half-way between "mountain man" and "mountain".
             for index, fragment_weight in enumerate(weights):
                 if fragment_weight < 1:
-                    fragments_without_this = fragments[:index] + fragments[index + 1:]
-                    weights_without_this = weights[:index] + weights[index + 1:]
-                    tokens, per_token_weights = self.get_token_ids_and_expand_weights(fragments_without_this,
-                                                                                      weights_without_this)
+                    fragments_without_this = fragments[:index] + fragments[index+1:]
+                    weights_without_this = weights[:index] + weights[index+1:]
+                    tokens, per_token_weights = self.get_token_ids_and_expand_weights(fragments_without_this, weights_without_this, device=device)
                     embedding_without_this = self.build_weighted_embedding_tensor(tokens, per_token_weights)
 
                     embeddings = torch.cat((embeddings, embedding_without_this.unsqueeze(0)), dim=1)
@@ -98,25 +112,22 @@ class EmbeddingsProvider:
                     #        inf at PI/2
                     # -> tan((1-weight)*PI/2) should give us ideal lerp weights
                     epsilon = 1e-9
-                    fragment_weight = max(epsilon, fragment_weight)  # inf is bad
+                    fragment_weight = max(epsilon, fragment_weight) # inf is bad
                     embedding_lerp_weight = math.tan((1.0 - fragment_weight) * math.pi / 2)
                     # todo handle negative weight?
 
                     per_embedding_weights.append(embedding_lerp_weight)
 
-            lerped_embeddings = self.apply_embedding_weights(embeddings, per_embedding_weights, normalize=True).squeeze(
-                0)
+            lerped_embeddings = self.apply_embedding_weights(embeddings, per_embedding_weights, normalize=True).squeeze(0)
 
-            # print(f"assembled tokens for '{fragments}' into tensor of shape {lerped_embeddings.shape}")
+            #print(f"assembled tokens for '{fragments}' into tensor of shape {lerped_embeddings.shape}")
 
             # append to batch
-            batch_z = lerped_embeddings.unsqueeze(0) if batch_z is None else torch.cat(
-                [batch_z, lerped_embeddings.unsqueeze(0)], dim=1)
-            batch_tokens = tokens.unsqueeze(0) if batch_tokens is None else torch.cat(
-                [batch_tokens, tokens.unsqueeze(0)], dim=1)
+            batch_z = lerped_embeddings.unsqueeze(0) if batch_z is None else torch.cat([batch_z, lerped_embeddings.unsqueeze(0)], dim=1)
+            batch_tokens = tokens.unsqueeze(0) if batch_tokens is None else torch.cat([batch_tokens, tokens.unsqueeze(0)], dim=1)
 
         # should have shape (B, 77, 768)
-        # print(f"assembled all tokens into tensor of shape {batch_z.shape}")
+        #print(f"assembled all tokens into tensor of shape {batch_z.shape}")
 
         if should_return_tokens:
             return batch_z, batch_tokens
@@ -150,10 +161,9 @@ class EmbeddingsProvider:
         for token_ids in token_ids_list:
             # trim eos/bos
             token_ids = token_ids[1:-1]
-
-            ## pad for textual inversions with vector length >1
-            # token_ids = self.textual_inversion_manager.expand_textual_inversion_token_ids_if_necessary(token_ids)
-
+            # pad for textual inversions with vector length >1
+            if self.textual_inversion_manager is not None:
+                token_ids = self.textual_inversion_manager.expand_textual_inversion_token_ids_if_necessary(token_ids)
             # truncate if necessary to max_length-2 (leaving room for bos/eos)
             token_ids = token_ids[0:self.max_token_count - 2]
             # add back eos/bos if requested
@@ -164,7 +174,7 @@ class EmbeddingsProvider:
 
         return result
 
-    def get_token_ids_and_expand_weights(self, fragments: list[str], weights: list[float]) -> (torch.Tensor, torch.Tensor):
+    def get_token_ids_and_expand_weights(self, fragments: list[str], weights: list[float], device: str) -> (torch.Tensor, torch.Tensor):
         '''
         Given a list of text fragments and corresponding weights: tokenize each fragment, append the token sequences
         together and return a padded token sequence starting with the bos marker, ending with the eos marker, and padded
@@ -204,23 +214,25 @@ class EmbeddingsProvider:
             all_token_ids = all_token_ids[0:max_token_count_without_bos_eos_markers]
             per_token_weights = per_token_weights[0:max_token_count_without_bos_eos_markers]
 
-        # pad out to a self.max_length-entry array: [eos_token, <prompt tokens>, eos_token, ..., eos_token]
+        # pad out to a self.max_length-entry array: [bos_token, <prompt tokens>, eos_token, pad_token...]
         # (typically self.max_length == 77)
         all_token_ids = [self.tokenizer.bos_token_id] + all_token_ids + [self.tokenizer.eos_token_id]
         per_token_weights = [1.0] + per_token_weights + [1.0]
         pad_length = self.max_token_count - len(all_token_ids)
-        all_token_ids += [self.tokenizer.eos_token_id] * pad_length
+        all_token_ids += [self.tokenizer.pad_token_id] * pad_length
         per_token_weights += [1.0] * pad_length
 
-        all_token_ids_tensor = torch.tensor(all_token_ids, dtype=torch.long, device=self.device)
-        per_token_weights_tensor = torch.tensor(per_token_weights, dtype=torch.float32, device=self.device)
+        all_token_ids_tensor = torch.tensor(all_token_ids, dtype=torch.long, device=device)
+        per_token_weights_tensor = torch.tensor(per_token_weights,
+                                                dtype=self.get_dtype_for_device(self.text_encoder.device),
+                                                device=device)
         #print(f"assembled all_token_ids_tensor with shape {all_token_ids_tensor.shape}")
         return all_token_ids_tensor, per_token_weights_tensor
 
 
     def build_weighted_embedding_tensor(self, token_ids: torch.Tensor, per_token_weights: torch.Tensor) -> torch.Tensor:
         """
-        Build a tensor that embeds the passed-in token IDs and applyies the given per_token weights
+        Build a tensor that embeds the passed-in token IDs and applies the given per_token weights
         :param token_ids: A tensor of shape `[self.max_length]` containing token IDs (ints)
         :param per_token_weights: A tensor of shape `[self.max_length]` containing weights (floats)
         :return: A tensor of shape `[1, self.max_length, token_dim]` representing the requested weighted embeddings
@@ -230,14 +242,13 @@ class EmbeddingsProvider:
         if token_ids.shape != torch.Size([self.max_token_count]):
             raise ValueError(f"token_ids has shape {token_ids.shape} - expected [{self.max_token_count}]")
 
-        z = self.text_encoder.forward(input_ids=token_ids.unsqueeze(0),
-                                      return_dict=False)[0]
+        z = self.text_encoder(token_ids.unsqueeze(0), return_dict=False)[0]
         empty_token_ids = torch.tensor([self.tokenizer.bos_token_id] +
-                                       [self.tokenizer.pad_token_id] * (self.max_token_count - 2) +
-                                       [self.tokenizer.eos_token_id], dtype=torch.int,
-                                       device=self.device).unsqueeze(0)
-        empty_z = self.text_encoder(input_ids=empty_token_ids).last_hidden_state
-        batch_weights_expanded = per_token_weights.reshape(per_token_weights.shape + (1,)).expand(z.shape)
+                                       [self.tokenizer.eos_token_id] +
+                                       [self.tokenizer.pad_token_id] * (self.max_token_count - 2),
+                                       dtype=torch.int, device=z.device).unsqueeze(0)
+        empty_z = self.text_encoder(empty_token_ids).last_hidden_state
+        batch_weights_expanded = per_token_weights.reshape(per_token_weights.shape + (1,)).expand(z.shape).to(z)
         z_delta_from_empty = z - empty_z
         weighted_z = empty_z + (z_delta_from_empty * batch_weights_expanded)
 
